from pathlib import Path
from junit_xml import TestSuite, TestCase


class LineReader(object):
    def __init__(self, file):
        self.current_line = ''
        self.current_line_number = 0
        self.file = file

    def get(self, line):
        while self.current_line_number != line:
            self.current_line_number = self.current_line_number + 1
            self.current_line = self.file.readline()

        return self.current_line


class Reporter(object):
    @staticmethod
    def create(reporter_type: str, files_with_info: list):
        if reporter_type == 'console':
            return ConsoleReporter(files_with_info)
        elif reporter_type == 'junit':
            return JUnitReporter(files_with_info)
        return None


class ConsoleReporter(Reporter):
    def __init__(self, files_with_info: list):
        self.files_with_info = files_with_info
        self.violations = 0

    def generate_report(self) -> str:
        output = ''
        for file, violations in self.files_with_info:
            output += self.__generate_file_output(file, violations)

        return output + '\n' + self.__create_summary()

    def __create_summary(self) -> str:
        files_number = len(self.files_with_info)
        return "Scanned %d file(s). Found %d issues.\n" % (files_number, self.violations)

    def __generate_file_output(self, file: Path, violations: list) -> str:
        output = ''
        if violations:
            with file.open() as cmake_file:
                line_reader = LineReader(cmake_file)
                output += "\n>>>" + str(file) + '<<<\n'
                output += self.__generate_file_violations(line_reader, violations)
        return output

    def __generate_file_violations(self, line_reader: LineReader, violations: list):
        output = ''
        for (violation_type, line_number) in violations:
            self.violations = self.violations + 1
            output += self.__generate_new_violation_prefix(line_number, violation_type)
            output += line_reader.get(line_number)
        return output

    @staticmethod
    def __generate_new_violation_prefix(line_number: int, violation_type: str):
        return "line: %4s %20s> " % (line_number, violation_type)


class JUnitReporter(Reporter):
    def __init__(self, files_with_info: list):
        self.files_with_info = files_with_info

    def generate_report(self) -> str:
        test_suites = self.__generate_test_suites()
        if not test_suites:
            test_suites = self.__generate_empty_test_suites()
        return TestSuite.to_xml_string(test_suites, prettyprint=True)

    def __generate_test_suites(self) -> list:
        test_suites = []
        for file_with_info in self.files_with_info:
            test_suites.append(self.__create_test_suite_for(file_with_info))
        return test_suites

    def __create_test_suite_for(self, file__with_info) -> TestSuite:
        test_case_list = self.__create_test_cases(file__with_info[0], file__with_info[1])
        return TestSuite(file__with_info[0], test_case_list)

    def __create_test_cases(self, file: Path, violations: list) -> list:
        test_cases = []
        with file.open() as cmake_file:
            line_reader = LineReader(cmake_file)

            for (violation, line_number) in violations:
                test_case = TestCase(name=violation, line=line_number)
                test_case.add_failure_info(message=self.__generate_failure_info(line_number, violation, line_reader))
                test_cases.append(test_case)
        return test_cases

    @staticmethod
    def __generate_failure_info(line_number: int, violation_type: str, line_reader: LineReader) -> str:
        return "line: %4s %20s> %s" % (line_number, violation_type, line_reader.get(line_number))

    @staticmethod
    def __generate_empty_test_suites() -> list:
        return [TestSuite('cmake-checker', [TestCase(name='No cmake files found')])]



import argparse
import sys
from pathlib import Path


def file_or_dir(path: str) -> Path:
    new_path = Path(path)
    if not new_path.exists():
        msg = "%s doesn't exist!" % path
        raise argparse.ArgumentTypeError(msg)
    return new_path


def parse_arguments() -> argparse.Namespace:
    arguments_parser = argparse.ArgumentParser()

    arguments_parser.add_argument('PATH',
                                  type=file_or_dir,
                                  nargs='+',
                                  help='Path to the file or directory where the checks should be done'
                                  )
    arguments_parser.add_argument('--warn-only',
                                  action='store_true',
                                  help='Program will return 0 even if violations are found'
                                  )
    arguments_parser.add_argument('--reporter',
                                  choices=['console', 'junit'],
                                  default='console',
                                  help='Specify type of reporter to output'
                                  )
    arguments_parser.add_argument('-o',
                                  '--output-file',
                                  type=argparse.FileType('w'),
                                  default=sys.stdout,
                                  help='Output results to file with given name'
                                  )
    arguments_parser.add_argument('--whitelist',
                                  type=argparse.FileType('r'),
                                  help='Whitelist file with rules to ignore certain files or dirs (.gitignore style)'
                                  )
    return arguments_parser.parse_args()



import sys

from cmake_checker.components.file_finder import provide_files_for_verification
from cmake_checker.components.parse_arguments import parse_arguments
from cmake_checker.components.verifier import Verifier
from cmake_checker.components.reporter import Reporter


def compute_exit_code(violations: list, warn_only: bool) -> int:
    if warn_only is True:
        return 0
    if any(len(v) for f, v in violations):
        return -1
    return 0


def main():
    arguments = parse_arguments()
    verify = Verifier()

    files = provide_files_for_verification(arguments.PATH, arguments.whitelist)
    files_with_info = verify.check_path(files)
    reporter = Reporter.create(arguments.reporter, files_with_info)
    arguments.output_file.write(reporter.generate_report())

    sys.exit(compute_exit_code(files_with_info, arguments.warn_only))


main()



from pathlib import Path


def provide_files_for_verification(paths: list, whitelist_content) -> list:
    file_finder = FileFinder(whitelist_content)

    file_list = []
    for path in paths:
        file_list += file_finder.get_all_cmake_files(path)
    return file_list


class FileFinder(object):
    def __init__(self, whitelist=None):
        if whitelist is None:
            self.ignore_patterns = []
        else:
            self.ignore_patterns = [line.rstrip('\n') for line in whitelist]

    def get_all_cmake_files(self, path: Path) -> list:
        if not path.is_dir():
            return [path]
        return self.__get_all_files_from_dir(path)

    def __get_all_files_from_dir(self, path: Path) -> list:
        all_cmake_files = sorted(path.glob('**/*.cmake')) + sorted(path.glob('**/CMakeLists.txt'))
        return [file for file in all_cmake_files if not self.__is_on_whitelist(file)]

    def __is_on_whitelist(self, file: Path) -> bool:
        return any(file.match(path_pattern) for path_pattern in self.ignore_patterns)



from pathlib import Path
from .lexer import Lexer


class Verifier(object):
    def __init__(self):
        self.lexer = Lexer()

    def check_path(self, files: list) -> list:
        files_with_issues = []
        for file in files:
            files_with_issues += [(file, self.__validate_file(file))]
        return files_with_issues

    def __validate_file(self, file: Path) -> list:
        with file.open() as cmake_file:
            return self.__find_issues(cmake_file.read())

    def __find_issues(self, data: str) -> list:
        return [(issue.type, issue.lineno) for issue in self.lexer.analyze(data)]



from setuptools import find_packages, setup

setup(
    name='cmake_checker',
    version='0.1.0',
    author='Maciej Patro',
    author_email='maciejpatro@gmail.com',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    url='https://github.com/MaciejPatro/cmake-checker',
    license='LICENSE',
    description='cmake-checker is a tool to search for violations to \'modern\' cmake rules.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    python_requires='>=3.5.0',
    install_requires=[
        "ply == 3.11",
        "junit-xml == 1.8",
    ],
    include_package_data=True,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)



import ply.lex as lex


class Lexer(object):

    states = (
        ('bracketcomments', 'exclusive'),
        ('declfunc', 'inclusive'),
        ('targetsources', 'exclusive'),
        ('setfunc', 'exclusive'),
        ('disabled', 'exclusive'),
        ('comments', 'exclusive'),
    )

    tokens = ['FILE_GLOB',
              'ADD_COMPILE_OPTIONS',
              'INCLUDE_DIRECTORIES',
              'LINK_DIRECTORIES',
              'LINK_LIBRARIES',
              'COMPILE_FLAGS',
              'CLOSING_COMMAND_WITH_CLAUSE',
              'ADD_DEFINITIONS',
              'ADD_COMPILE_DEFINITIONS',
              'MODIFY_ENV_VARIABLE',
              'CACHE_IN_SET',
              'PARENT_DIR_ACCESS',
              'ENDFUNCTION',
              'PARENT_SCOPE'
              ]

    t_FILE_GLOB = r'file[ \t]*\([ \t]*GLOB'
    t_ADD_COMPILE_OPTIONS = r'add_compile_options[ \t]*\('
    t_ADD_DEFINITIONS = r'add_definitions[ \t]*\('
    t_ADD_COMPILE_DEFINITIONS = r'add_compile_definitions[ \t]*\('
    t_INCLUDE_DIRECTORIES = r'(^include_directories[ \t]*\()|([ \t]+include_directories[ \t]*\()'
    t_LINK_DIRECTORIES = r'(^link_directories[ \t]*\()|([ \t]+link_directories[ \t]*\()'
    t_LINK_LIBRARIES = r'(^link_libraries[ \t]*\()|([ \t]+link_libraries[ \t]*\()'
    t_COMPILE_FLAGS = r'(CMAKE_CXX_FLAGS)|(CMAKE_C_FLAGS)'
    t_CLOSING_COMMAND_WITH_CLAUSE = r'(endif\(.+\))|(endmacro\(.+\))|(endforeach\(.+\))'
    t_setfunc_MODIFY_ENV_VARIABLE = r'[ \t]*ENV\{'
    t_setfunc_CACHE_IN_SET = r'[ \t]+CACHE'
    t_targetsources_PARENT_DIR_ACCESS = r'\.\.\/\.\.'

    def __init__(self):
        self.lexer = lex.lex(module=self)
        self.is_in_function = False
        self.bracket_comment_count = 0

    @staticmethod
    def t_ANY_error(t: lex.Token) -> None:
        t.lexer.skip(1)

    @staticmethod
    def t_INITIAL_disabled_setfunc_targetsources_bracketcomments_newline(t: lex.Token) -> None:
        r"""\n+"""
        t.lexer.lineno += len(t.value)

    def t_begin_bracketcomments(self, t: lex.Token) -> None:
        r"""(([^\\]\#)|.{0}\#)(\[=*\[)"""
        self.bracket_comment_count = t.value.count('=')
        t.lexer.push_state('bracketcomments')

    def t_bracketcomments_end(self, t: lex.Token) -> None:
        r"""\]=*\]"""
        if self.bracket_comment_count is t.value.count('='):
            t.lexer.pop_state()

    def t_begin_declfunc(self, t: lex.Token) -> None:
        r"""[ \t]*function[ \t]*\("""
        t.lexer.push_state('declfunc')
        self.is_in_function = True

    def t_declfunc_ENDFUNCTION(self, t: lex.Token) -> lex.Token:
        r"""endfunction\(.+\)"""
        t.lexer.pop_state()
        self.is_in_function = False
        return t

    def t_declfunc_end(self, t: lex.Token) -> None:
        r"""endfunction\([ \t]*\)"""
        t.lexer.pop_state()
        self.is_in_function = False

    def t_setfunc_PARENT_SCOPE(self, t: lex.Token):
        r"""[ \t]+PARENT_SCOPE"""
        if not self.is_in_function:
            return t
        return None

    @staticmethod
    def t_begin_targetsources(t: lex.Token) -> None:
        r"""[ \t]*target_sources[ \t]*\("""
        t.lexer.push_state('targetsources')

    @staticmethod
    def t_targetsources_end(t: lex.Token) -> None:
        r"""\)"""
        t.lexer.pop_state()

    @staticmethod
    def t_begin_setfunc(t: lex.Token) -> None:
        r"""set[ \t]*\("""
        t.lexer.push_state('setfunc')

    @staticmethod
    def t_setfunc_end(t: lex.Token) -> None:
        r"""\)"""
        t.lexer.pop_state()

    @staticmethod
    def t_comments_begin_disabled(t: lex.Token) -> None:
        r"""cmake-check[ \t]+disable"""
        t.lexer.push_state('disabled')

    @staticmethod
    def t_disabled_end(t: lex.Token) -> None:
        r"""cmake-check[ \t]+enable"""
        t.lexer.pop_state()

    @staticmethod
    def t_INITIAL_targetsources_setfunc_begin_comments(t: lex.Token) -> None:
        r"""([^\\]\#)|.{0}\#"""
        t.lexer.push_state('comments')

    @staticmethod
    def t_comments_end(t: lex.Token) -> None:
        r"""\n+"""
        t.lexer.pop_state()
        t.lexer.lineno += len(t.value)

    def analyze(self, data: str) -> list:
        self.lexer.input(data)
        self.lexer.lineno = 1
        self.is_in_function = False

        tokens = []

        while True:
            token = self.lexer.token()
            if not token:
                break
            tokens.append(token)

        return tokens
